/* Generated by Cython 3.0.10 */

#ifndef __PYX_HAVE__recometrics__cpp_funs
#define __PYX_HAVE__recometrics__cpp_funs

#include "Python.h"

#ifndef __PYX_HAVE_API__recometrics__cpp_funs

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
    #define __PYX_EXTERN_C extern "C++"
#endif

#ifndef DL_IMPORT
  #define DL_IMPORT(_T) _T
#endif

__PYX_EXTERN_C double ddot_(int const *, double const *, int const *, double const *, int const *);
__PYX_EXTERN_C float sdot_(int const *, float const *, int const *, float const *, int const *);

#endif /* !__PYX_HAVE_API__recometrics__cpp_funs */

/* WARNING: the interface of the module init function changed in CPython 3.5. */
/* It now returns a PyModuleDef instance instead of a PyModule instance. */

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC initcpp_funs(void);
#else
/* WARNING: Use PyImport_AppendInittab("cpp_funs", PyInit_cpp_funs) instead of calling PyInit_cpp_funs directly from Python 3.5 */
PyMODINIT_FUNC PyInit_cpp_funs(void);

#if PY_VERSION_HEX >= 0x03050000 && (defined(__GNUC__) || defined(__clang__) || defined(_MSC_VER) || (defined(__cplusplus) && __cplusplus >= 201402L))
#if defined(__cplusplus) && __cplusplus >= 201402L
[[deprecated("Use PyImport_AppendInittab(\"cpp_funs\", PyInit_cpp_funs) instead of calling PyInit_cpp_funs directly.")]] inline
#elif defined(__GNUC__) || defined(__clang__)
__attribute__ ((__deprecated__("Use PyImport_AppendInittab(\"cpp_funs\", PyInit_cpp_funs) instead of calling PyInit_cpp_funs directly."), __unused__)) __inline__
#elif defined(_MSC_VER)
__declspec(deprecated("Use PyImport_AppendInittab(\"cpp_funs\", PyInit_cpp_funs) instead of calling PyInit_cpp_funs directly.")) __inline
#endif
static PyObject* __PYX_WARN_IF_PyInit_cpp_funs_INIT_CALLED(PyObject* res) {
  return res;
}
#define PyInit_cpp_funs() __PYX_WARN_IF_PyInit_cpp_funs_INIT_CALLED(PyInit_cpp_funs())
#endif
#endif

#endif /* !__PYX_HAVE__recometrics__cpp_funs */
